; ****************************
; ZX80 Breakout - Main Program
; ****************************
; (c)1980 Macronics
;
; RAM required: 1K
;
; Disassembled by Paul Farrow, www.fruitcake.plus.com
; Initial version 11th March 2011 (revised 16th April 2011)


; =====
; Notes
; =====

; This disassembly listing uses tab characters with tab stops at every 8th column.

; The program spans from $4000 (16384) to $40EA (16618) inclusive.
;
; The Display file is located at $40FF (16639). It begins with a newline character ($76), followed by 24 rows of
; 31 columns. Each column is terminated with a newline character ($76).
;
; The following program variables are used:
;
;               $40F8   = Holds a sequence counter which indicates what action to perform at each frame.
;                         The bit combinations have the following meanings:
;                           00 = Move the ball.
;                           01 = Move the bat.
;                           10 = Delay.
;                           11 = Move the bat.
;               $40F9   = Holds the character representation of the number of balls ('0' to '9').
;               $40FA   = Top of stack. Not used.
;               $44FB   = Not used.
;               $44FC   = Not used.
; IY+$22        $40FD   = Holds the low byte of the display file address of the bat.
; IY+$23        $40FE   = Holds the number of scan lines to produce. It is used by the ROM display routines, which addresses this
;                         location using IY+$23.


; ====================
; Assembler Directives
; ====================

; For use with the TASM assembler.

#define DEFB .BYTE
#define DEFW .WORD
#define DEFM .TEXT
#define DEFS .FILL
#define EQU  .EQU
#define ORG  .ORG
#define END  .END

; ======================================================================================================================================================

        ORG     $4000

; --------
; Draw Bat
; --------
; This routine is used to erase the bat at its old position and draw it at its new position.
; The bat is 5 characters wide.
; Entry: HL=Display file address of the centre of the bat.
;        D=Character to use to draw the bat.
; Exit : HL=Display file address of the centre of the bat.
; Time : 93 T-states.

L4000:  LD      (HL),D                  ; 7     Draw the centre character.
        DEC     HL                      ; 6
        LD      (HL),D                  ; 7     Draw the left characters of the bat.
        DEC     HL                      ; 6
        LD      (HL),D                  ; 7

        INC     HL                      ; 6     Advance to the right hand character position of the bat.
        INC     HL                      ; 6
        INC     HL                      ; 6

        LD      (HL),D                  ; 7     Draw the right characters of the bat.
        INC     HL                      ; 6
        LD      (HL),D                  ; 7

        DEC     HL                      ; 6     Point back to the centre of the bat.
        DEC     HL                      ; 6
        RET                             ; 10

; -------------------
; Program Entry Point
; -------------------

L400E:  LD      DE,$0020                ; The number of characters per row, including the line terminator.

        LD      HL,$4100                ; Start of the display file.

; Initialise the top four rows to black.

        LD      A,$80                   ; Black character.
        LD      B,A                     ; Fill 4 rows (128/32=4)

L4017:  LD      (HL),A                  ; Fill a byte.
        INC     HL
        DJNZ    L4017                   ; Repeat for all characters in the top four rows of the display file.

; Initilaise the next four rows with brick characters ('*').

        LD      B,A                     ; B holds $80.

L401C:  LD      (HL),$14                ; '*'.
        INC     HL
        DJNZ    L401C                   ; Repeat for all characters in the next four rows of the display file.

; Initialise the rest of the screen to black.

L4021:  LD      (HL),A                  ; Fill 2 bytes (2*256=512=16*32).
        INC     HL
        LD      (HL),A
        INC     HL
        DJNZ    L4021                   ; Repeat for all remaining characters of the display file.

; Insert a newline character at the start of the display file and at the end of each row, i.e. as the 32nd character of each row.

        LD      HL,$40FF                ; Point to the byte before the display file.
        LD      B,$19                   ; 25 characters.

L402C:  LD      (HL),$76                ; Insert the row terminator.
        ADD     HL,DE                   ; Advance to the next row.
        DJNZ    L402C                   ; Repeat for all rows.

; Initialise the bat and ball locations.
; The ball is initially at location (17,0) and moving down and to the right.
; The centre of the bat is initially at location (23,6), and hence the bat spans locations (23,4) to (23,8).

        DEC     H                       ; Set the initial ball display file location to 17,0;
        INC     L

        LD      C,$01                   ; BC holds the initial horizontal direction offset of the ball, i.e. $0001.
                                        ; DE holds the initial vertical direction offset of the ball, i.e. $0020.

        LD      IY,$40DB                ; Base address of program variables.
        LD      (IY+$22),$E6            ; Set the display location of the bat to row 23, column 4.

        LD      SP,$40FA                ; Set up a stack at 16634 (+250).
        EXX

; HL'=Display file location of the ball.
; DE'=Vertical direction offset ($0020 or $FFE0).
; BC'=Horizontal direction offset ($0001 or $FFFF).

        LD      HL,$2500                ; H holds '9' which is the number of balls. L holds the frame sequence counter.
        PUSH    HL                      ; Save the frame sequence counter.

; A PAL TV picture consists of 25 frames per second, consisting of 625 scan lines. Each frame is divided into 2 interlaced fields, referred to
; as odd and even fields, with each field consisting of 312.5 scan lines. A field occupies 20ms and hence each scan line takes 64us.
; To produce a steady TV picture, the display routine within the ZX80 ROM has been carefully structured so that each iteration of the loop
; takes exactly 20ms to execute. The ZX80 is driven by a 6.5MHz ceramic resonator/filter, which is halved and tuned down to approximately 3.21MHz.
; Each iteration of display loop within the ROM takes 64167 T-states and this spans 20ms. Therefore each T-states lasts 311.67ns, and hence the
; CPU clock frequency assumed by the ZX80 is 3.20835MHz (which is close to the 3.2MHz marked on the ZX80 schematic). Ceramic resonators do not
; have a tight tolerance and so some variation around this figure is to be expected. The ZX80 ROM outputs 303 scan lines (56 in the top border,
; 192 in the main area and 55 in the bottom border), with each requiring 207 T-states to generate. Therefore 303 * 207 = 62721 T-states are
; required to span these scan lines, with an additional 27 T-states used to return control to the display loop. The remaining 1419 T-states
; are used to read the keyboard and update the FRAMES system variable. Note that with a clock frequency of 3.20835MHz, each scan line takes 64.52us.
; This means that the TV field produced by the ZX80 spans 309.99 of these slightly longer scan lines.
;
; The Breakout game contains a display loop that outputs 303 scan lines per TV field (56 in the top border, 192 in the main display area,
; and 55 in the bottom border). This is identical to the standard ZX80 TV frame. To achieve a steady display, the functionality performed at
; each field must be structured to fill this time as accurately as possible. The game code is divided into four sections, and these are cycled
; through one per TV field. It takes 67 T-states to run the sequence control loop, leaving 1363 T-states (0.425ms) for each game section.
; The game sections actually take the following durations (taking into account the time to dispatch to the game section routine):
;
; Part 1 = 1165 T-states (Range 1132 to 1264 T-states) - Move the ball.
; Part 2 = 1154 T-states (Range 1154 to 1182 T-states) - Move the bat.
; Part 3 = 1165 T-states                               - Delay.
; Part 4 = 1154 T-states (Range 1154 to 1182 T-states) - Move the bat.
;
; The bat can therefore move twice as fast as the ball.
;
; There is a range of values for each section due to small differences that can occur between different paths within those routines. The values
; are all close enough to the target value of 0.418ms that any difference simply causes the next field to be displayed slightly sooner or later
; than the desired 20ms rate. Although it is desirable to produce TV fields at a precise rate of 20ms, a small amount of variation is tolerable
; and television sets can cope with this. The target duration of 0.425ms is based on the CPU speed assumed by the ZX80 ROM, but this may vary
; slightly between machines due to tolerance of the resonator and so is only a guide to the time available.
;
; The Breakout game is written to generate a 50Hz display field rate suitable for the UK. The ZX80 ROM generates a 60Hz display field rate
; by reducing the number of top and bottom border lines by 24 each. The game can be modified to support a 60Hz display rate by modifying the
; following line as indicated.

L4045:  LD      (IY+$23),$38            ; 19    There will be 56 scan lines above the ZX80 TV display area to achieve a 50Hz field rate.
                                        ;       [Change this to 32 ($20) scan lines for USA 60Hz]

        POP     HL                      ; 10    Fetch the byte at $40F8, which is used as a sequence counter to determine what action to perform this frame.
        INC     L                       ; 4     Increment the counter.
        PUSH    HL                      ; 11    Store the updated counter.

        IN      A,($FE)                 ; 11    Turn on the vertical sync generator. It be will be turned off approximately 6.5 scan lines from now.
                                        ;       [Any I/O read to an address with A0 low will turn on the vertical sync generator]

; The sequence counter determines the action to perform on this TV field.

        LD      A,L                     ; 4     Fetch the sequence counter.
        AND     $03                     ; 7     Keep only the lower 2 bits, forming 4 possibilities to consider.
        JR      Z,L40A4                 ; 12/7  Sequence index 0: Move the ball (typically 1142 T-states, though ranges from 1109 to 1241 T-states).

        CP      $02                     ; 7
        JR      NZ,L4073                ; 12/7  Sequence index 1 and 3: Read the keyboard and move the bat (typically 1117 T-states, though ranges from 1117 to 1145 T-states).

; Sequence index 2: Delay.

        LD      B,$57                   ; 7     Delay for 7+(87*13)-5=1133 T-states.

L4059:  DJNZ    L4059                   ; 13/8

; Generate the top TV border and main area.

L405B:  OUT     ($FE),A                 ; 11    Turn off the vertical sync generator.
                                        ;       [Note that the ZX80 ROM uses an OUT ($FF),A instruction to do this, but any I/O OUT will do]

; The first scan line produced by the program is too short but this does not cause a problem for TV sets since all subsequent scan lines are produced correctly.
; The first scan line is too short by 14 T-states. The problem can be resolved by loading a value of $E9 into the A register instead of $EC. This will introduce
; an additional 12 T-states, and so reduces the error to -2 T-states. The A register hold a counter that is incremented until it reaches 0, at which point the start
; of the next scan line is triggered. Each decrement of the value loaded into the A register introduces a 4 T-states delay, and hence decrement the value by 3
; introduces a 12 T-states delay.

        LD      A,$EC                   ; 7     Timing counter for the time left to generate for the first scan line of the top border. Thereafter set to $DD by the ROM.
        LD      B,$19                   ; 7     There are 25 HALTs within the display file, i.e. 24 rows and 1 top border.
        LD      HL,$C0FF                ; 10    Point to the start of the display file, i.e. the initial HALT representing the top border, but with bit 15 of the address set.
        CALL    $01AD                   ; 17    DISP-2. Produce the top border and main area.

; Generate the bottom TV border.

        LD      A,$F0                   ; 7     Timing counter for the time left to generate for the first scan line of the bottom border. Thereafter set to $DD by the ROM.
        INC     B                       ; 4     Increment the row count, i.e. B holds $01.
        DEC     HL                      ; 6     Decrement the screen address, i.e. point to the last HALT.
        DEC     (IY+$23)                ; 23    Decrement the blank line counter, i.e. there will be 55 scan lines below the ZX80 TV display area.

        CALL    $01AD                   ; 17    DISP-2. Produce the bottom border.

        JR      L4045                   ; 12    Repeat the main program loop for the current ball.

; ------------
; Move the Bat
; ------------
; This routine reads the keybaord and moves the bat. It can move from column 2 to column 28 on row 23.
; It is moved to the right by key 8 or 3, and moved to the left by key 5 or 6.
; The routine takes between 1117 T-states and 1145 T-states, typically 1143 T-states.

L4073:  LD      L,(IY+$22)              ; 19    Fetch the address of the bat.
        LD      H,$43                   ; 7

; Erase the bat at the current location.

        LD      D,$80                   ; 7     Black character.
        CALL    L4000                   ; 17+93 Erase the bat.

; Upper byte selects keyboard row to read.
;
;          Bit0  Bit1  Bit2  Bit3  Bit4    Bit4  Bit3  Bit2  Bit1  Bit0
;          ----  ----  ----  ----  ----    ----  ----  ----  ----  ----
; $F7FE =    1     2     3     4     5       6     7     8     9     0   = $EFFE
; $FBFE =    Q     W     E     R     T       Y     U     I     O     P   = $DFFE
; $FDFE =    A     S     D     F     G       H     J     K     L    N/L  = $BFFE
; $FEFE =  SHIFT   Z     X     C     V       B     N     M     .   SPACE = $7FFE

        LD      B,$E7                   ; 7     Select both halves of the top keyboard row (1-5 and 6-0).
        LD      C,$FE                   ; 7     The keyboard I/O port.
        LD      A,L                     ; 4     Fetch the low byte of the laser base display file address.
        IN      D,(C)                   ; 12    Read the keyboard.

; Test the left key.

        BIT     2,D                     ; 8     Was key 8 (or 3) pressed?

        JR      NZ,L408D                ; 12/7  Jump ahead if not.

; Key 8 (or 3) was pressed so attempt to move to the right.

        CP      $FC                     ; 7     Is the centre of the bat at column 28, hence the right of the bat at column 30?
        JR      Z,L408D                 ; 12/7  Jump ahead if so.

        INC     L                       ; 4     Increment the bat display file address, i.e. one cell to the right.

; Test the right key.

L408D:  BIT     4,D                     ; 8     Was key 5 (or 6) pressed?
        JR      NZ,L4096                ; 12/7  Jump ahead if not.

; Key 5 (or 6) was pressed so attempt to move to the left.

        CP      $E2                     ; 7     Is the centre of the bat at column 2, hence the left of the bat at column 0?
        JR      Z,L4096                 ; 12/7  Jump ahead if so.

        DEC     L                       ; 4     Decrement the bat display file address, i.e. one cell to the left.

; Draw the bat at the new location.

L4096:  LD      D,$03                   ; 7     Graphic consisting of top white with bottom black.
        CALL    L4000                   ; 17+93 Draw the bat.

        LD      (IY+$22),L              ; 19    Store the new bat display file address low byte.

; Delay to balance the other frame sequence routines.

        LD      B,$3A                   ; 7     Perform a delay of 7+(58*13)-5=756 T-states.

L40A0:  DJNZ    L40A0                   ; 13/8  Delay to balance other frame sequence routines.

        JR      L405B                   ; 12    Jump back into the main loop to display the next TV field.

; -------------
; Move the Ball
; -------------
; This routine moves the ball. The ball can only move in four directions: up right, up left, down right, down left.
; When the left or right edge of the screen is reached, the horizontal direction of the ball is reversed.
; When the top or bottom edge of the screen is reached, the vertical direction of the ball is reversed.
; If the bat is hit then the vertical direction of the ball is reversed.
; If a brick is hit then the vertical direction of the ball is reversed.
; The routine takes between 1033 T-states and 1241 T-states, typically 1142 T-states.

L40A4:  EXX                             ; 4     Fetch the location and direction of the ball.

        LD      (HL),$80                ; 10    Erase the ball.

        ADD     HL,BC                   ; 11    Add the horizontal direction offset ($0001 or $FFFF).
        ADD     HL,DE                   ; 11    Add the vertical direction offset ($0020 or $FFE0).

; HL holds the address of the new ball location.

        ADD     HL,BC                   ; 11    Add the horizontal direction offset again to allow examination of the character horizontally after the new position.

; Apply the horizontal direction offset.

        LD      A,(HL)                  ; 7     Fetch the character horizontally after the new location.
        AND     A                       ; 4
        SBC     HL,BC                   ; 15    Point back to the new location.
        CP      $76                     ; 7     Is it the left or right edge of the screen?
        JR      NZ,L40B9                ; 12/7  Jump ahead if not.

; The left or right edge of the screen has been reached, so reverse the horizontal direction.

        LD      A,B                     ; 4     Invert the high byte of the horizontal direction offset (either $FF or $00).
        CPL                             ; 4
        LD      B,A                     ; 4

        LD      A,C                     ; 4     Invert the low byte of the horizontal direction offset (either $FF or $01).
        CPL                             ; 4
        INC     A                       ; 4
        LD      C,A                     ; 4

; Apply the vertical direction offset.

L40B9:  LD      A,(HL)                  ; 7     Fetch the character at the new location.
        ADD     HL,DE                   ; 11    Add the vertical direction offset again to allow examination of the character vertically after the new position.
        CP      $14                     ; 7     Has a brick character been hit?
        JR      Z,L40D8                 ; 12/7  Jump ahead if so.

; A brick character has not been hit.

        LD      A,(HL)                  ; 7     Fetch the character vertically after the new location.
        CP      $03                     ; 7     Is the bat?
        JR      Z,L40D8                 ; 12/7  Jump ahead if so.

The new location vertically does not contain the bat.

        LD      A,H                     ; 4     Fetch the high byte display file address of the character vertically after the new location.
        CP      $40                     ; 7     Has the ball reached the top of the screen?
        JR      Z,L40D8                 ; 12/7  Jump ahead if it has.

; The ball has not reached the top of the screen.

        AND     A                       ; 4
        SBC     HL,DE                   ; 15    Point back to the new location.
        CP      $44                     ; 7     Has the bottom of the screen been reached?
        JR      NZ,L40E0                ; 12/7  Jump ahead if it has not.

; The bottom of the screen has been reached and hence a ball has been lost.

        POP     AF                      ; 10    Fetch the ball number.
        DEC     A                       ; 4     Decrement it.
        PUSH    AF                      ; 11    Save the new ball number.

        CP      $1B                     ; 7     Has the ball number character reached '.', i.e. all balls lost?
        JP      Z,L400E                 ; 10/10 Jump if so to restart the game.

; Join here when either:
; - A brick has been hit.
; - The bat has been hit.
; - The top of the screen has been hit and there is at least one ball left.

L40D8:  LD      A,D                     ; 4     Invert the high byte of the vertical direction offset (either $FF or $00)
        CPL                             ; 4
        LD      D,A                     ; 4

        LD      A,E                     ; 4     Invert the low byte of the vertical direction offset (either $FF or $01).
        CPL                             ; 4
        INC     A                       ; 4
        LD      E,A                     ; 4

        ADD     HL,DE                   ; 11    Point back to the new ball location.

; Join here when the ball has not hit any other character or a screen boundary.

L40E0:  POP     AF                      ; 10    Fetch the ball number character.
        PUSH    AF                      ; 11    Save it again.

        LD      (HL),A                  ; 7     Draw the ball number character in the display file.

        EXX                             ; 4     Save the ball location and direction.

        LD      B,$45                   ; 7     Perform a delay of 7+(69*13)-5=899 T-states.

L40E6:  DJNZ    L40E6                   ; 13/8  Delay to balance other frame sequence routines.

        JP      L405B                   ; 10    Jump back into the main loop to display the next TV field.

; ======================================================================================================================================================

; ===========
; FLOW CHARTS
; ===========

; ------------
; Move the Bat
; ------------
; $4073. The routine takes between 1117 T-states and 1145 T-states, although it is typically 1143 T-states.
;
;
;                                                   $4073                                MOVE BAT
;                                                     |
;                                                  26 |
;                                                     v
;                                      Erase Bat  117 |
;                                                     v
;                                  Read keyboard   30 |
;                                                     v
;                                                   8 |
;                                                     v         12
;                                                     +-----------------------------------------------+
;                                                   7 |                                               |
;                                  Key 8 Pressed      v                                               |
;                                                   7 |                                               |
;                                                     v         12                                    |
;                                                     +-----------------------+                       |
;                                                   7 |     Bat At Column 28  |                       |
;                                                     v                       |                       |
;                                                   4 |                       |                       |
;                                                     v                       v                       |
;                                               $408D +<----------------------+<----------------------+
;                                                     |
;                                                   8 |
;                                                     v         12
;                                                     +-----------------------------------------------+
;                                                   7 |                                               |
;                                  Key 5 Pressed      v                                               |
;                                                   7 |                                               |
;                                                     v         12                                    |
;                                                     +-----------------------+                       |
;                                                   7 |     Bat At Column 2   |                       |
;                                                     v                       |                       |
;                                                   4 |                       |                       |
;                                                     v                       v                       |
;                                               $4096 +<----------------------+<----------------------+
;                                                     |
;                                       Draw Bat  117 |
;                                                     v
;                                                  26 |
;                                                     v
;                                           (58*13)-5 |
;                                                     v
;                                                  12 |
;                                                     v
;                                           JUMP BACK TO MAIN LOOP


; -------------
; Move the Ball
; -------------
; $40A4. The routine takes between 1033 T-states and 1241 T-states, although it is typically 1142 T-states.
;
;
;                                                   $40A4                                MOVE BALL
;                                                     |
;                                                   4 |
;                                                     v
;                                      Erase Bat   10 |
;                                                     v
;                                                  66 |
;                                                     v         12
;                                                     +-----------------------+
;                                                   7 |   Not edge of screen  |
;                                                     v                       |
;                                                  28 |                       |
;                                                     v                       |
;                                               $40B9 +<----------------------+
;                                                     |
;                                                  25 |
;                                                     v         12
;                                                     +-----------------------------------------------+
;                                                   7 |   Brick hit                                   |
;                                                     v                                               |
;                                                  14 |                                               |
;                                                     v         12                                    v
;                                                     +---------------------------------------------->+
;                                                   7 |   Bat hit                                     |
;                                                     v                                               |
;                                                  11 |                                               |
;                                                     v         12                                    v
;                                                     +---------------------------------------------->+
;                                                   7 |   Top of screen hit                           |
;                                                     v                                               |
;                                                  26 |                                               |
;                                       12            v                                               |
;     +-----------------------------------------------+                                               |
;     |                 Bottom of screen not hit    7 |                                               |
;     |                                               v                                               |
;     |                                            32 |                                               |
;     |                                               v         10                                    |
;     |                                               +-------------------> $400E Restart Game        |
;     |                                            10 |   All balls lost                              |
;     |                                               v                                               |
;     |                                         $40D8 +<----------------------------------------------+
;     |                                               |
;     |                                            39 |
;     |                                               v
;     +---------------------------------------------->+ $40E0
;                                                     |
;                                                  21 |
;                                                     v
;                                       Draw Bat    7 |
;                                                     v
;                                                  11 |
;                                                     v
;                                           (69*13)-5 |
;                                                     v
;                                                  10 |
;                                                     v
;                                           JUMP BACK TO MAIN LOOP


; ======================================================================================================================================================

        END

